#define _GNU_SOURCE /* strchrnul関数使用のために必要 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define MAIN_C
#include "exitfail.h"

/* 表示する文字数 */
#define DISP_WIDTH  60 /* 幅 */
#define DISP_HEIGHT 17 /* 高さ */

/* CSVデータ各要素のサイズ */
#define ZIPCODE_LEN 8
#define STREET_LEN  512
#define CITY_LEN    64
#define PREF_LEN    16
#define FLAG_NUM    6
/* CSVデータ1行の要素数 */
#define COLUMN_NUM  (5 + FLAG_NUM)

/* CSVデータ構造体 */
typedef struct {
	long code;                 /* 全国地方公共団体コード */
	char zipcode[ZIPCODE_LEN]; /* 郵便番号 */
	char street[STREET_LEN];   /* 町域名 */
	char city[CITY_LEN];       /* 市区町村名 */
	char pref[PREF_LEN];       /* 都道府県名 */
	long flag[FLAG_NUM];       /* フラグ配列 */
} csvline_t;

#define BASENAME(p) ((strrchr((p), '/') ? : ((p) - 1)) + 1)

/**
 * 行表示関数
 * @param pcsvline 表示するCSVデータ構造体へのポインタ(NULLの場合は表示済データの総数を表示)
 */
static void printline(csvline_t *pcsvline)
{
	static int count = 0;	/* 表示したデータ総数 */
	static int line = 0;	/* 一画面中に表示した行数 */
	int disp_width = DISP_WIDTH, disp_height = DISP_HEIGHT, newline;
	char buf[1024];

	/* CSVデータの各要素を表示 */
	if (pcsvline) {
		/* 郵便番号が入っていない場合は表示しない */
		if (!pcsvline->zipcode[0])
			return;

		/* 各要素を表示フォーマットに展開 */
		snprintf(buf, sizeof(buf),
			 "%6d %05ld %s %s %s %s %ld %ld %ld %ld %ld %ld",
			 count++, pcsvline->code, pcsvline->zipcode,
			 pcsvline->street, pcsvline->city, pcsvline->pref,
			 pcsvline->flag[0], pcsvline->flag[1],
			 pcsvline->flag[2], pcsvline->flag[3],
			 pcsvline->flag[4], pcsvline->flag[5]);
	}
	/* CSVデータの総数を表示 */
	else {
		/* データ総数を表示フォーマットに展開 */
		sprintf(buf, "Count: %6d", count);
	}

	/* 今回追加される表示行数を計算 */
	newline = (strlen(buf) + (disp_width - 1)) / disp_width;
	/* 1画面を超える場合、入力があるまで一時停止 */
	if (line + newline >= disp_height) {
		getchar();
		/* 表示行数を初期化 */
		line = 0;
	}
	/* 実際に表示する */
	printf("%s\n", buf);
	/* 表示行数を更新 */
	line += newline;
}

/**
 * main関数
 * @param argc 引数なしの場合はusage表示のみ
 * @param argv 第1引数として読み込みCSVファイル名を指定
 * @return exit値
 */
int main(int argc, char *argv[])
{
	FILE *pcsvfile; /* CSVファイルポインタ */
	csvline_t csvline; /* CSVデータ */
	char buf[256], *pbuf, *pcol[COLUMN_NUM];
	int i;

	exitfail_init();

	/* 引数が指定されなかった場合、usage表示して終了 */
	if (argc < 2) {
		printf("Usage: %s <csvfile>\n", BASENAME(argv[0]));
		return EXIT_SUCCESS;
	}

	/* CSVファイルオープン */
	pcsvfile = fopen(argv[1], "r");
	if (!pcsvfile)
		exitfail_errno("fopen");

	/* CSVデータを初期化 */
	memset(&csvline, 0, sizeof(csvline));
	/* CSVファイルから1行読み込む */
	while (fgets(buf, sizeof(buf), pcsvfile)) {
		/* 各要素へのポインタ配列を初期化 */
		memset(pcol, 0, sizeof(pcol));
		/* 次のカンマ(または行末)を見つけてトークン化 */
		pbuf = strtok(buf, ",\n");
		for (i = 0; i < COLUMN_NUM && pbuf; i++) {
			/* 前後のダブルクォートは除去する */
			if (*pbuf == '"')
				*strchrnul(++pbuf, '"') = '\0';
			/* 要素へのポインタを保持 */
			pcol[i] = pbuf;
			/* 次のカンマ(または行末)を見つけてトークン化 */
			pbuf = strtok(NULL, ",\n");
		}
		/* 要素数が不足している場合、次行にスキップ */
		if (i < COLUMN_NUM)
			continue;

		/* 新しいデータの場合 */
		if (strcmp(pcol[1], csvline.zipcode) ||
		    strtol(pcol[8], NULL, 10)) {
			/* 保持済みのデータを表示 */
			printline(&csvline);

			/* CSVデータ各要素を保持 */
			memset(&csvline, 0, sizeof(csvline));
			csvline.code = strtol(pcol[0], NULL, 10);
			strncpy(csvline.zipcode, pcol[1],
				sizeof(csvline.zipcode) - 1);
			strncpy(csvline.street, pcol[2],
				sizeof(csvline.street) - 1);
			strncpy(csvline.city, pcol[3],
				sizeof(csvline.city) - 1);
			strncpy(csvline.pref, pcol[4],
				sizeof(csvline.pref) - 1);
			for (i = 0; i < FLAG_NUM; i++)
				csvline.flag[i] = strtol(pcol[5 + i], NULL, 10);
		}
		/* 既存データへの追加の場合 */
		else
			/* 町域名の続きを追加 */
			strncat(csvline.street, pcol[2],
				(sizeof(csvline.street) -
				 strlen(csvline.street)) - 1);
	}
	/* 保持済みのデータを表示 */
	printline(&csvline);

	/* CSVファイルクローズ */
	fclose(pcsvfile);

	/* データ総数を表示 */
	printline(NULL);

	return EXIT_SUCCESS;
}
